      program driver
c**********************************************************************
c                   SPLIB Copyright (C) 1995                         **
c                     Indiana University                             **
c**********************************************************************
      implicit none
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c
c  Sample driver for SPLIB, sparse matrix iterative library.  This sets
c  up the storage for A,x,b, and reads in a data file that drives the choice
c  of methods and various print/graphing options.
c
c  This version allocates all needed space from a single integer 
c  array "work", and SPLIB then uses a memory manager to allocate all
c  additional array storage needed.  This version also reads in the 
c  transpose of the matrix from a Harwell/Boeing format file.  The
c  reason for "transpose" is because H/B matrices are in compressed
c  sparse column format, while this driver treats them as in compressed
c  sparse row format.
c
c  This driver also reads in a sequence of solvers/preconditioners from
c  an input file called "methods" applying them until solver = 0 is
c  specified.  The xgraph files showing convergence history for each
c  method on a single run are written to the same file for comparisons.
c
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c  Memory management module, which sets up the main array work from
c  which all vectors with problem dependent sizes are drawn.
c  Only exception is xstar, the true solution vector.  This is rarely 
c  available and so if you want true error norms, you will need to
c  modify the code.
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

c     ------------------------
c     Total memory for arrays:
c     ------------------------
      integer     mtot
      parameter ( mtot =  36 000 000)                                  
      integer  work
      common /memory/ work(mtot)                                    
      include 'include/memory.inc'

c     --------------------------------------------------
c     Number of integer words per double precision word:
c     --------------------------------------------------
      include 'include/intdbl.inc'

c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c        Documentation
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c
c  Solve a general sparse system Ax = b by means of calls to 
c  various iterative methods and preconditioners.  The driver will
c  loop over an input data file until it encounters a sentinal value,
c  and then quit.  The input matrix is read by Sparskit readmt.f,
c  modified to correctly handle the single right hand side case, and
c  modified to use pointers in a work array to define arrays read in.
c  A should be in the new Harwell/Boeing format, with right hand sides
c  getting a special format line of their own (the five line header
c  version of H/B format, not the four line header).
c
c  Because most of the preconditioners rely on the diagonal of the matrix
c  being present, a routine matchk then checks the data structure,
c  modifying it by inserting diagonal entries if necessary.  Those 
c  entries are set to zero; all that changes is the data structure,
c  not the matrix.
c
c------------------------------------------------------------------
c    
c     Randall Bramley
c     Department of Computer Science
c     Indiana University, Bloomington
c     bramley@cs.indiana.edu
c     Thu Jul 13 12:58:23 EST 1995
c
c------------------------------------------------------------------
c
c  Parameters:
c  ===========
c
c  mtot       : maximum number of integer words allowed for all problem
c               dependent arrays.  This includes the matrix, the vectors
c               defining the right hand side and solution, the preconditioners,
c               and the additional work vectors needed.
c
c  intdbl     : number of integer words per double precision word for
c               target machine. 
c
c  Variables (also see the descriptions in include file instr.inc):
c  ==============================================================
c
c  work    : Array of work storage, for preconditioners, solvers, and.
c            instrumentation routines.  The matrices and vectors used
c            are all indexed from this integer array, using pointers.
c            The following are all just integer pointers into work():
c
c  n       : Order of matrix A.
c  A       : (int pointer) real*8 array for nonzero entries of 
c                          matrix A; of length nnz = rwptr(n+1)-1.
c  colind  : (int pointer) integer array of column indices for
c                          entries in A; of length nnz = rwptr(n+1)-1.
c  rwptr   : (int pointer) integer array of pointers to beginning position
c                          of rows in arrays A and colind.  Length n+1.
c
c  x       : (int pointer) real*8 vector of unknowns to be solved for.
c  g       : (int pointer) real*8 vector of initial estimate of x, if 
c                          provided by readmt3.
c  xs      : (int pointer) real*8 vector of correct solution, if
c                          provided by readmt3.
c  b       : (int pointer) real*8 right hand side vector of system to
c                          be solved: Ax = b
c            See lrhs and lsol for options on this vector.
c
c  next    : points to next available location in array work()
c  remain  : number of integer words remaining in work() to be used.
c
c  tol     : tolerance on norm used in stopping test.
c  ireltv  : variable to indicate whether stopping test
c            is relative to size of right hand side vector or not.
c            (0 for not, 1 to use relative residual).
c  rpcprm  : real preconditioner parameter -- 
c             - tolerance for numerical dropping with ILUT;
c             - omega parameter for SSOR(omega);
c             - relaxation constant for MILU.
c             - sparsity pattern indicator for ECIMGS
c  maxits  : Maximum number of iterations allowed.
c            On return from splib.f, actual number of iterations.
c  solmeth : Solution method to use.
c             solmeth = 1  ---> BiCG
c             solmeth = 2  ---> CG on normal equations AA'y = b, x = A'y.
c             solmeth = 3  ---> CG on normal equations A'A x = A'b
c             solmeth = 4  ---> CGS
c             solmeth = 5  ---> CGSTAB
c             solmeth = 6  ---> GMRES(k) 
c             solmeth = 7  ---> transpose-free QMR
c             solmeth = 8  ---> Bi-CG stabilized (Templates version of CG-Stab).
c             solmeth = 9  ---> Templates version of GMRES (still under work)
c             solmeth = 10 ---> Jacobi
c             solmeth = 11 ---> Gauss-Seidel
c             solmeth = 12 ---> SOR(omega)
c             solmeth = 13 ---> Orthomin(k)
c  krylov  : Size of Krylov subspace to use in GMRES or Orthomin; k above.
c  pcmeth  : Preconditioning method to use.
c             pcmeth = 0  ---> no preconditioner.
c             pcmeth = 1  ---> ILU(s)
c             pcmeth = 2  ---> MILU(s,rpcprm)
c             pcmeth = 3  ---> ILUT(s,rpcprm)
c             pcmeth = 4  ---> SSOR(rpcprm)
c             pcmeth = 5  ---> TRID(s), where s is the block size.
c             pcmeth = 6  ---> ILU(-1), the space saver version of ILU(0)
c             pcmeth = 7  ---> ECIMGS.
c  ipcprm  : integer preconditioning parameter, usually levels of fill.
c            For tridiag preconditioning, it means the block size.
c  n       : Order of matrix A, must be less equal to nmax.
c  nnz     : Number of nonzero entries in A, must be less equal 
c            to nzmax.
c  ind     : Integer used for loops and indexing.
c  nrhs    : Integer contains number of right hand sides returned by
c            readmt3.  Although multiple right hand sides may be read in,
c            currently the code will only solve the system for the first one.
c  job     : Integer to indicate what is to be read in readmt3.
c            Here only used for job = 2, i.e. read matrix only
c            and job = 3, i.e. read matrix and 1 right hand side
c  lrhs    : Logical variable indicating whether or not the rhs is supplied
c            in readmt3.  If it is not, an artificial one is created in setb.
c  lsol    : Logical variable indicating whether or not the true
c            solution vector is provided by readmt3.
c  lgus    : Logical variable indicating whether or not an initial guess for
c            solution vector x is provided by readmt3.
c
c  ierr    : Integer variable indicating error conditions.
c             ierr = -m --> Not enough workspace provided; m more double
c                           precision words needed in array work().
c             ierr = 0  --> Successful return; tolerance achieved.
c             ierr = 1  --> The initial guess satisfies the stopping test.
c                           This is often a clue that rhs not set properly.
c             ierr = 2  --> Tolerance reset 8 times in attempt to
c                           get true residual small enough.
c             ierr = 3  --> Convergence not achieved in maxits iterations.  
c             ierr = 4  --> Initial preconditioned residual is more than
c                           20 orders of magnitude larger than initial 
c                           unpreconditioned residual.  Not exactly an error
c                           condition, but when this happens the iterations never
c                           converge in practice.  Try another preconditioner.
c             4<ierr<9  --> breakdown condition of attempt to divide by
c                           near-zero number in algorithm.
c             Note that any of conditions 0, 1, 2 may mean success.
c
c  title   : Character array to hold title of Harwell/Boeing matrix.
c  key     : Character array for short title of Harwell/Boeing matrix.
c  guesol  : Character array for readmt.  Has interpretation
c            guesol(1:1) = 'G' or 'g' --> initial guess provided
c            guesol(2:2) = 'X' or 'x' --> correct solution provided.
c  res     : Final residual norm (2-norm)
c  err     : Final error norm (2-norm)
c  firstmat: Logical variable indicating if first time to execute splib,
c            so that matrix is only read in on first pass.  In common block
c            of instr.inc
c  fmtd    : Integer indicating if the matrix input file is formatted or not.
c              fmtd = 0 ---> formatted (ascii) input file
c              fmtd = 1 ---> unformatted (binary) input file
c
c  I/O Files
c  =========
c
c  methods : Input file for indicating method to be used, 
c            partitioning, etc.  This file has the repeated structure
c     
c     solmeth		
c     krylov
c     pcmeth
c     levelsfill
c     accel_parm
c     tol  ireltv
c     maxits
c     instlvl
c     iredo
c     
c     An example file "methods" to compare GMRES(15) and 
c     CGSTAB, both with MILU preconditioning might look like:
c     
c     5		!   Use CGSTAB iterative method.
c     0	        !   Not used for CGSTAB ...
c     2		!   Use MILU preconditioning method
c     1		!   Number of levels of fill in preconditioner
c     1.0d0	!   Relaxation parameter for MILU.
c     1.0d-8 0	!   Stopping criterion of || A*x - b || < 1.0e-8
c     500	!   Allow up to 500 iterations
c     5         !   Provide full instrumentation
c     0         !   Scaling method (0 = none)
c     1         !   Input matrix is in formatted (ascii) file.
c     0         !   Recompute preconditioner 
c     
c     6		!   Use GMRES iterative method.
c     15	!   Use 15 Krylov vectors for GMRES 
c     2		!   Use MILU preconditioning method
c     1		!   Number of levels of fill in preconditioner
c     1.0d0	!   Relaxation parameter for MILU.
c     1.0d-8 0	!   Stopping criterion of || A*x - b || < 1.0e-8
c     500	!   Allow up to 500 iterations
c     5         !   Provide full instrumentation
c     0         !   Scaling method (0 = none)
c     1         !   Input matrix is in formatted (ascii) file.
c     1         !   Use same preconditioner as on previous run.
c     
c     0		!   Solmeth = 0 indicates stop run.
c
c  matrix  : Input file containing Harwell/Boeing matrix.  Actually
c            this collection solves a system with the transpose of
c            the H/B matrix provided, but a short routine to read 
c            in a H/B matrix and output its transpose is also 
c            available.  If fmtd = 0, then the file is assumed to be
c            in a binary file (Fortran unformatted).
c
c  results : Output file containing summary of results from run.
c
c  resi    : Output file with residual norm vs. iteration number.
c  rest    : Output file with residual norm vs. elapsed time.
c
c  reli    : Output file with relative residual norm vs. iteration number.
c  relt    : Output file with relative residual norm vs. elapsed time.
c
c  Mresi   : Output file with preconditioned residual norm vs. iteration
c            number.
c  Mrest   : Output file with preconditioned residual norm vs. elapsed
c            time.
c
c  Mreli   : Output file with preconditioned relative residual norm vs.
c            iteration number.
c  Mrelt   : Output file with preconditioned relative residual norm vs.
c            elapsed time.
c
c  oetli   : Output file with Oettli and Prager norm vs. iteration number.
c  oetlt   : Output file with Oettli and Prager norm vs. elapsed time.
c
c  errori  : Output file with error norm vs. iteration number
c            (if lsol = true).
c  errort  : Output file with error norm vs. time (if lsol = true).
c
c   ===================
c   Files opened/used:
c   ===================
c
c  The I/O unit numbers are provided in the vector iunit(20).  Note that units
c  4-6 are mandatory, and unit 20 can be changed to be the same as unit 6.
c  The other units are optional, depending on the error norm(s) you wish to
c  plot/use.
c  
c  For this driver the following assignments are made to iunit():
c
c  4 --> iunit(1): Input file for matrix in Harwell/Boeing format
c  5 --> iunit(2): Input file with control parameters
c  6 --> iunit(3): Output file with echoed input and summary of run(s).
c  
c  7 --> iunit(4): Output unit for || Ax - b ||  vs. iter
c  8 --> iunit(5): Output unit for || inv(M) (Ax-b) || 
c                  vs. iter
c  9 --> iunit(6): Output unit for || Ax - b || / || r_0 ||
c                  vs. iter
c 10 --> iunit(7): Output unit for || inv(M) (Ax-b) || / 
c                                  || inv(M) r_0 || vs. iter
c 11 --> iunit(8): Output unit for || Ax - b ||_inf /
c                                ( ||A||_inf ||x||_1 + ||b||_inf) vs. iter
c 12 --> iunit(9): Output unit for || x - xstar || vs. iter
c 13 --> iunit(10): Not used.
c  
c 14 --> iunit(11): Output unit for || Ax - b || vs. time
c 15 --> iunit(12): Output unit for || inv(M) (Ax-b) || vs. time
c 16 --> iunit(13): Output unit for || Ax - b || / || r_0 || 
c                   vs. time
c 17 --> iunit(14): Output unit for || inv(M) (Ax-b) || / 
c                   || inv(M) r_0 || vs. time
c 18 --> iunit(15): Output unit for || Ax - b ||_inf / 
c                   ( ||A||_inf ||x||_1 + ||b||_inf) vs. time
c 19 --> iunit(16): Output unit for || x - xstar || vs. time
c
c
c  =================
c  External Routines
c  =================
c
c  readmt3 : Reads Harwell/Boeing matrix into a, colind, rwptr format
c  bmux    : Sparse matrix multiply (Sparskit)
c  dnrm2   : Blas routine for two-norm of vector
c
c  ==================
c  Workspace Required
c  ==================
c
c  The amount of workspace depends on the methods used, the levels
c  of fill-in (and way of defining fill-in levels), and the number
c  of Krylov subspace vectors used.  The following guide excludes 
c  the storage for x, r, and b which appear in the calling
c  sequence for all the routines.  Note that GMRES(k) depends
c  on the size k of the Krylov subspace.  Also note that the methods
c  with fill-in will require more storage for alu, jlu, ju.  The size
c  shown is for the preconditioners with no fill-in allowed, and does
c  not include the storage that depends on nnz.
c
c     Routine		Reals		Integers
c     -------		-----		--------
c
c     (m)ilus 		n          	2n
c     ilut    		2n + 1		3n
c     ssor    		1		n (but ju is recycled for this)
c     trid    		4n		n
c     ilu0    		n		n
c     ecimgs  		?		?
c
c     Routine		Reals		
c     -------		-----		
c     bicg		6n              }
c     cgne		3n              }
c     cgnr		3n              }  Add 2n for all of these if
c     cgs 		6n              }  the instrumentation routine
c     cgstab 		6n              }  is to be used at all; the
c     gmres(k)		k*(n+k+5)+n + 1 }  extra is for work arrays
c     qmrtf   		10n             }  used in residual computation.
c     bicgs	 	6n              }
c
c  ==========================
c  Memory Management Options:
c  ==========================
c
c  Either the memory manager can be initialized in this driver, or it can
c  be initialized in the splib routine.  If it is initialized in this driver,
c  comment out the block of code that sets "next" before each of the calls
c  to addtbl in this routine, uncomment out the call to initbl in this 
c  driver, and call splib with work(1).  Within splib.f, simply comment out
c  the call to initbl. 
c
c  If the splib.f routine initializes the memory manager (as would occur for
c  applications other than cadyf), comment out the calls to initbl and
c  addtbl in this driver, and uncomment out the pointer setting blocks
c  involving "next" that precedes the calls to addtbl.  Also, uncomment
c  out the call to initbl inside of splib.f.
c
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c
c   Declarations of variables:
c
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
      include 'include/symbols.inc'


c     --------------------------------
c     Double precision array pointers:
c     --------------------------------
      integer a, x, b, xs, g, Dl, Dr

c     -------------------------
c     Double precision scalars:
c     -------------------------
      real*8 tol, rpcprm
      real*8 err

c     ----------------------------
c     Function programs from BLAS:
c     ----------------------------
      real*8 dnrm2
      external dnrm2

c     -------------------------------------
c     Function programs from memory manager:
c     -------------------------------------
      integer addtbl, loctbl, sizeof, storof, getnext, memleft
      logical istbl
      external addtbl, loctbl, sizeof, storof, istbl, getnext, memleft

c     -----------------------
c     Integer array pointers:
c     -----------------------
      integer rwptr, colind
      integer next, remain

c     ---------
c     Integers:
c     ---------
      integer ierr, maxits, iparms(2), n, nnz
      integer i, k, mneed
      integer job, nrhs, ipcprm
      integer pcmeth, solmeth, krylov, ndiag
      integer digits
      integer fmtd
      integer scale

c     ------------------
c     Logical variables:
c     ------------------
      logical chkdiag
      logical lsol, lrhs, lgus, firstmat
      integer ireltv, iredo

c     --------------------
c     Character variables:
c     --------------------
      character type*3, guesol*2
      character title*72, key*8
      character solver*5, precond*4
      character from*8

c     --------------------------
c     Instrumentation variables:
c     --------------------------
      integer iunit(16), instlvl
      common /truesol/ xstar
      real*8 xstar(1)
      data firstmat /.true./

c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c      Beginning of Executable Statements:
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
c-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

c     ----------------------------------------------
c     Input file for matrix in Harwell/Boeing format
c     ----------------------------------------------
      iunit(3) = 4

c     ----------------------------------
c     Input file with control parameters
c     ----------------------------------
      iunit(2) = 5

c     ----------------------------------------------------
c     Output file with echoed input and summary of run(s).
c     ----------------------------------------------------

      iunit(1) = 60

c     ------------------------------------------------
c     Other unit number somewhat arbitrarily assigned:
c     ------------------------------------------------
      do i = 4,20
         iunit(i) = i+3
      end do 

c     -----------------------------------------
c     Open required files for input and results
c     -----------------------------------------
      open(iunit(2),file='methods')
      open(iunit(1),file='results.html')

c     ----------------------------------------------------------------
c     Initialize memory management.  Have it write to 'results' file.
c     ----------------------------------------------------------------
      call initbl(iunit(1),mtot)

c     -------------------------------------------------
c     Set pointer to next available location in work():
c     -------------------------------------------------
        next = 1
        remain = mtot

c*****************************************************************************
c****** Big "go to" loop over problem solution/preconditioner specifications:*
c*****************************************************************************

  20  continue
      ierr = 0

c     -----------------------------------------------------------------
c     Read in parameters that define stopping tolerances, methods to be
c     used.  Unpack information into pcmeth, ipcprm, etc.  Open 
c     additional I/O files if specified by instlvl.
c     -----------------------------------------------------------------
      call unpack (ireltv,   rpcprm    ,pcmeth  ,
     1             ipcprm   ,solmeth ,krylov  ,
     2             tol     ,maxits  , iredo,
     3             iunit   ,
     4             instlvl , scale, fmtd,
     5             title, key, solver  , precond)

c     ------------------------------------------------------
c     Try to read right hand side and solution vector.
c     If not supplied, create the right hand side vector b.
c     ------------------------------------------------------
      job = 3

c     -----------------------------------------------------------------------
c     Read matrix and vectors, and set up pointers.  Note that readmt3 will
c     insert extra positions after the colind array, to provide for 
c     the possible expansion of data structure by inserting missing diagonal
c     entries.
c     -----------------------------------------------------------------------
      if (firstmat) then

c       ------------------------------------------------------------
c       Force matrix data structure to have diagonal entries present
c       ------------------------------------------------------------
        chkdiag = .true.
        if (fmtd .eq. 0) then
           open(iunit(3),file='matrix',form='unformatted')
           call readunf2(mtot,n,n,nnz,a,colind,rwptr,iunit(3),
     1       work(next),mneed,ierr,nrhs,b,title,key,type,digits)
        else
           open(iunit(3),file='matrix',form='formatted')
           call readmt3(mtot  ,job,iunit(3),a,colind,rwptr,b,xs,g,nrhs,
     1        work(next),mneed,guesol,n,n,nnz,title,key,type,ierr,
     2        chkdiag)
        end if
        if (ierr .ne. 0) then
             write(iunit(1),5000) ierr
             write(iunit(1),5010) mneed
             stop
        end if 
        lrhs = (job .ge. 3)
        lsol = (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')
        lgus = (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g')

c       --------------------------------------------------------------
c       Now update next and remain to reflect storage used in readmt3.
c       --------------------------------------------------------------
        from = 'readmt'
        k = addtbl('rwptr', n+1, 'integer', from)
        k = addtbl('colind', nnz, 'integer', from)
        k = addtbl('a', nnz, 'real', from)
        if (lrhs) k = addtbl('b', n*nrhs, 'real', from)
        if (lgus) k = addtbl('g', n*nrhs, 'real', from)
        if (lsol) k = addtbl('xs', n*nrhs, 'real', from)
        next = getnext()
        remain = memleft()

c       --------------------------------------------------------------
c       Call matrix checker to insert missing diagonal entries, and to
c       sort rows into increasing column order:
c       --------------------------------------------------------------
        if (remain .lt. n) then
             write(iunit(1),5050) next
             stop
        end if
        call matchk(work(a),work(colind),work(rwptr),n,work(next),
     &                ndiag)

c       -----------------
c       Set up initial x.
c       -----------------
        from = 'driver'
        x = addtbl('x', n, 'real', from)
        remain = memleft()
        if (remain .le. 0) then
            next = getnext()
            write(iunit(1),5020) next
            stop
        end if

c       --------------------------------------------------------
c       If job < 3, the right hand side vector must be provided.
c       --------------------------------------------------------
        if (.not. lrhs) then
             from = 'driver'
             b = addtbl('b', n, 'real', from)
             remain = memleft()
             if (remain .le. 0) then
                next = getnext()
                write(iunit(1),5030) next
                stop
             end if
             call setb(n,work(b))
        end if

c  FORCE b to be all ones:
c             call setb(n,work(b))

c=======================================================================
c           End "if firstmat" clause 
c=======================================================================
      end if 


c     ----------------------------------------------------
c     Echo information about matrix and sol/precon methods
c     ----------------------------------------------------
      write(iunit(1),6070) title(1:72)
      write(iunit(1),6080) key

c     ---------------------------------------------------------
c     Copy over true solution vector to xstar, if it was
c     provided in readmt3.  If it was not, then we cannot
c     check the error norm, so set that in the errchk vector
c     in case the user tried to specify it.  Note: since
c     having a true solution vector is rare, in this driver
c     and in splib, it is passed using the common block truesol.
c     If you wish to see the true error norms, you must change
c     the declaration of the vector xstar in common truesol in
c     both driver.f and instr.f, recompile, and then uncomment 
c     out the next five lines.
c     ---------------------------------------------------------
c      if (lsol) then
c          call dcopy(n,work(xs),1,xstar(1),1)
c      else
c          instlvl  = min(instlvl, 5)
c      end if 
      instlvl  = min(instlvl, 5)

c     --------------------------------------------------------
c     Set up initial x appropriately.  If a "guess" vector was
c     provided in readmt3, use it.  Otherwise, set x = 0.
c     --------------------------------------------------------
      if (lgus) then
         call dcopy(n,work(g),1,work(x),1)
      else
         call setconst('dbl',n,work(x),work(x),0,1.0d0)
      end if

c     -----------------------------------
c     Pack up the iparms vector for splib
c     -----------------------------------
      iparms(1) = ipcprm
      iparms(2) = krylov

c     ------------------------------------------
c     Set pointer to scaling vectors for system.
c     ------------------------------------------
      if (.not. istbl('Dl', from )) then
        Dl = addtbl('Dl', n, 'real', from)
      else
        Dl = loctbl('Dl', from)
      end if
      if (.not. istbl('Dr', from )) then
        Dr = addtbl('Dr', n, 'real', from)
      else
        Dr = loctbl('Dr', from)
      end if

c     ----------------------------
c     Call splib to apply methods.
c     ----------------------------
      remain = memleft()
      call splib(work(a), work(colind), work(rwptr), n, 
     1                 work(x), work(b), work(1), remain,
     2                 tol, ireltv, maxits, pcmeth, solmeth,
     3                 scale, work(Dl), work(Dr),
     4                 iparms, rpcprm, iredo, ierr, iunit, instlvl)

c      open(77,file='x')
c      call prtval(77,'dbl',n,work(x),work(x))

c     ---------------------------------------------
c     Compute error norm if true solution available
c     ---------------------------------------------
      if (lsol) then
          call dcopy(n, work(x), 1, work(next), 1)
          call daxpy(n, -1.0d0, work(xs),1, work(next),1)
          err = dnrm2(n,work(next),1)
          write(iunit(1),2010) err
      end if

c     -----------------------------------------------
c     Pick up next solver/preconditioner pair to try.
c     -----------------------------------------------
      firstmat = .false.

      go to 20

c========================================================================
c           Format Statements
c========================================================================

 6020  format('Order of matrix = ',i6,/,'Number of nonzeros ',i8)
 6040  format('Tolerance on residual = ',1pe10.4)
 6050  format('Iterative method used: ',a8,' with ',a8,
     +         ' preconditioning')
 6060  format('Levels of fill allowed in precond: ',i3)
 6070  format(/,40('--'),/,'Problem title: ',a71)
 6080  format('Problem key: ',a8)
 6090  format('Acceleration parameter: ',f8.2)

 2010  format('Final error 2-norm: ',1pe12.5)

 5000  format(' *** Error in readmt ; ierr = ',i4)
 5010  format(' *** Error in readmt ; mneed = ',i7)
 5020  format(' *** Error in driver; ran out of storage to set up x.',
     &        '     next = ',i7)
 5030  format(' *** Error in driver; ran out of storage to set up b.',
     &        '     next = ',i7)
 5040  format(' *** Error in driver; ran out of storage to set up r.',
     &        '     next = ',i7)
 5050  format(' *** Error in driver; insuff storage for matchk',
     &        '     next = ',i7)

	end
c----------------------------------------------------------------------
