c========================================================================        
c========================================================================        
c
c  Utility.f contains routines and functions that are generally useful for 
c  all codes, and need not be specific for a particular application like
c  Stokes solvers or domain decomposed block solvers.  Currently contains
c  the routines:
c      unpack  : reads in and unpacks control parameters
c      strings : sets up character strings that name the method 
c                and preconditioner used; for Xgraph output files
c
c      matchk  : checks matrix and inserts diagonal entries if needed.
c      readf   : read in a vector stored in ascii, one entry per row
c      readmt3 : version of readmt from SPARSKIT which uses pointers into
c                a work array, handles rectangular matrices, and provides
c                space for diagonal entry in each row if specified.
c      rdline  : reads in a single line of input for readmt3
c      bmux    : CSR A - multiply: y = A*x
c      btmux   : CSR A'- multiply: y = A'*x
c      btmuxp  : CSR A'- multiply: y = y + A'*x
c      btmuxm  : CSR A'- multiply: y = y - A'*x
c      setconst: Set an array to a constant value (int or dbl)
c      setb    : Set up a nontrivial rhs vector b.
c      getdia  : Sparskit routine to extract diagonal from a matrix.
c      prtval  : Print out a vector (int or dbl)
c      scala   : Scaling of matrix A and right hand side b
c      unscalx : Unscale solution x at end of computations.
c
c      double precision function dlamch  : routines from LAPACK 
c      subroutine dlamc1
c      subroutine dlamc2
c      double precision FUNCTION dlamc3
c      subroutine dlamc4
c      subroutine dlamc5
c
c========================================================================        

c**********************************************************************
c                   SPLIB Copyright (C) 1995                         **
c                     Indiana University                             **
c**********************************************************************

c========================================================================        

      subroutine strings(solmeth, solver, pcmeth, precond)
c-----------------------------------------------------------------------
c  Set up character strings identifying the solver/precond pair as
c  needed in SPLIB's Xgraph output files.
c-----------------------------------------------------------------------
      implicit none
      character solver*5, precond*4
      integer solmeth, pcmeth

      if( solmeth .eq. 1) then
         solver = 'BiCG '
      elseif (solmeth .eq. 2) then
         solver = 'CGNE '
      elseif (solmeth .eq. 3) then
         solver = 'CGNR '
      elseif (solmeth .eq. 4) then
         solver = 'CG-Sq'
      elseif (solmeth .eq. 5) then
         solver = 'CGStb'
      elseif (solmeth .eq. 6) then
         solver = 'GMRES'
      elseif (solmeth .eq. 7) then
         solver = 'QMRTF'
      elseif (solmeth .eq. 8) then
         solver = 'BiCGS'
      elseif (solmeth .eq. 9) then
         solver = 'GMRST'
      elseif (solmeth .eq. 10) then
         solver = 'JACOB'
      elseif (solmeth .eq. 11) then
         solver = 'GAUSS'
      elseif (solmeth .eq. 12) then
         solver = 'SOR'
      elseif (solmeth .eq. 13) then
         solver = 'ORTM'
      elseif (solmeth .eq. 14) then
         solver = 'CSCGS'
      end if
c
      if (pcmeth .eq. 1) then
        precond = 'ILU '
      elseif (pcmeth .eq. 2) then
        precond = 'MILU'
      elseif (pcmeth .eq. 3) then
        precond = 'ILUT'
      elseif (pcmeth .eq. 4) then
        precond = 'SSOR'
      elseif (pcmeth .eq. 5) then
        precond = 'TRID'
      elseif (pcmeth .eq. 6) then
        precond = 'ILU0'
      elseif (pcmeth .eq. 7) then
        precond = 'EMGS'
      elseif (pcmeth .eq. 0) then
        precond = 'NOPC'
      end if
c
	return
	end
c========================================================================

       subroutine matchk(a, ja, ia, n, diag, ndiag)
c------------------------------------------------------------------------
c
c  Check the matrix A in CSR format to make sure it includes all diagonal
c  entries.  If it does not have them, insert 0.0 into the data structure
c  for each missing diagonal entry.  At the same time, sort the rows of
c  A to have entries in increasing column order.  This is useful for 
c  preconditioning methods like SSOR or ILU(-1), and for matrix splitting
c  based iterative methods like Jacobi.
c
c  a, ja, ia	: matrix A in CSR format
c  n		: order of matrix A
c  diag		: integer work vector of length n.  On return, diag(1:n)
c                 contains pointers to the index of each diagonal entry
c                 in the modified CSR data structure.
c  ndiag	: integer, on return gives number of diag entries that
c                 were missing from original matrix data structure.
c  
c------------------------------------------------------------------------
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Thu Aug 25 08:06:44 EST 1994
c
c------------------------------------------------------------------------
       implicit none

c      ------------------------
c      Declaration of arguments
c      ------------------------
       real*8  a(*)
       integer n, ja(*), ia(n+1), diag(n), ndiag

c      ---------------
c      Local variables
c      ---------------
       integer row, lastptr, ibegin, iend, j, lp, ap, added

c      --------------------------------------------
c      Value to insert for missing diagonal entries
c      --------------------------------------------
       real*8 INSERT 
       parameter (INSERT = 0.0d0)

c=======================================================================
c          Beginning of Executable Statements
c=======================================================================


c      ----------------------------------------------------------------
c      Find diagonal entry for each row.  If it exists, set diag(row)
c      to be its index in CSR data structure.  Otherwise, set diag(row)
c      to be zero.
c      ----------------------------------------------------------------
       ndiag = 0
       do row = 1, n
          diag(row) = 0
          do j = ia(row), ia(row+1)-1
             if (ja(j) .eq. row) then
                diag(row) = j
                go to 20
             end if 
          end do
          ndiag = ndiag + 1
 20    continue
       end do

c      --------------------------------------------------------
c      If ndiag = 0 here, then all diagonal entries are present
c      in the data structure.  Sort the rows and exit.
c      --------------------------------------------------------
       if (ndiag .eq. 0) then
          do row = 1, n
             ibegin = ia(row)
             iend   = ia(row+1)
             call rowinc( iend - ibegin, ja(ibegin), a(ibegin) )
          end do
          return
       end if

c      ---------------------------------------------------------
c      Otherwise, insert some diagonal entries into data struct.
c      ---------------------------------------------------------
       lastptr = ia(n+1) - 1 + ndiag
       lp = lastptr
       ap = ia(n+1) - 1

       do row = n, 1, -1

          ibegin = ia(row)
          iend   = ia(row+1)

c         --------------------------------------------
c         Copy a, ja over, shifting to right as needed
c         --------------------------------------------
          do j = iend-1, ibegin, -1
             a(lp)  = a(ap)
             ja(lp) = ja(ap)
             lp = lp - 1 
             ap = ap - 1 
          end do 

c         --------------------------------
c         Insert diagonal entry if missing
c         --------------------------------
          added = 0
          if (diag(row) .eq. 0) then
             a(lp) = INSERT
             ja(lp) = row
             lp  = lp - 1
             added = 1
          end if 

c         ----------------------------
c         Sort newly modified row of A
c         ----------------------------
          call rowinc( iend - ibegin + added, ja(lp+1), a(lp+1) )
c  check for duplicates
c                if (iend - ibegin .gt. 1) then
c                do j = ibegin, iend -2
c                 if (ja(j) .eq. ja(j+1)) then
c                    write(*,*) 'squack: row ', row, ' col ', ja(j)
c                    Stop
c                 end if
c                end do 
c                end if
c  end check for duplicates

c         -------------------
c         Set new row pointer
c         -------------------
          ia(row+1) = lastptr + 1

c         -------------------
c         Change last pointer
c         -------------------
          lastptr = lp

       end do
c  begin check for duplicates
c       write(*,*) 'squack: alls ok i guess'
c  end check for duplicates
       return
       end

c
c========================================================================        
c
      subroutine readf(iounit,n,rarray)
c-----------------------------------------------------------------------
c  Read in a right hand side vector f.
c-----------------------------------------------------------------------
      integer iounit, n
      real*8 rarray(n)
      read(iounit,*) (rarray(i), i = 1, n)
      return
      end
c
c=======================================================================
c
      subroutine readmt3(maxstr,job,iounit,a,colind,rwptr,b,sol,g,nrhs,
     &         work,mneed,guesol,nrow,ncol,nnz,title,key,type,ierr,
     &         chkdiag)
c-----------------------------------------------------------------------
c
c This routine reads in the transpose of a H/B file, and is a modification
c of the original SPARSKIT routine readmt.  This one differs in that the
c storage vectors a, colind, rwptr, rhs are pointers into a singly
c dimensioned array work().  At each stage, readmt3 will check the amount
c of storage provided to see if it can read the next batch of data.
c Furthermore, it can be assured that the matrix data structure includes
c space for the diagonal entries by adjusting it as we go along, if the
c argument chkdiag = .true.
c
c The pointers will be aligned on double word boundaries, if intdbl = 2.
c For this to work, however, the first word of the integer array work must
c also be aligned on a double word boundary.
c
c Note that although this is used in SPLIB to read matrices stored in
c CSR format, it works equally well on standard H/B format matrices
c (which are in CSC format).
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University, Bloomington
c     Mon Aug  8 12:25:22 EST 1994
c
c-----------------------------------------------------------------------
c
c   Control Arguments:
c   ==================
c
c maxstr =  max storage provided for a, colind, rwptr, rhs.  This will 
c           possibly limit the amount of data read in.  It should
c           be at least  n  + 1  +  nnz  integers for rwptr and colind, 
c                        nnz   real*8 for a,
c                        n     real*8's for one right hand side b,
c                        n     real*8's for one solution vector sol
c
c job	 = integer to indicate what is to be read. (note: job is an
c          input and output parameter, it can be modified on return)
c          job = 0    read the values of ncol, nrow, nnz, title, key,
c                     type and return.  Matrix is not read and arrays
c                     a, colind, rwptr, rhs are not touched.
c          job = 1    read structure only, i.e., the arrays colind and rwptr.
c          job = 2    read matrix including values, i.e., a, colind, rwptr
c          job = 3    read matrix and right hand sides: a,colind,rwptr,rhs.
c		      rhs may contain initial guesses and exact 
c                     solutions appended to the actual right hand sides.
c		      this will be indicated by the output parameter
c                     guesol [see below]. 
c                     
c iounit = logical unit number from which to read the matrix.
c work   = integer array, used for storing the vectors returned.
c next   = integer pointer, indicating next available position in
c          array work.
c
c   ============== 
c   Return Values:
c   ============== 
c
c job    = on return job may be modified to the highest job it could
c          do: if job=2 on entry but no matrix values are available it
c          is reset to job=1 on return. Similarly if job=3 but no rhs 
c          is provided then it is reset to job=2 or job=1 depending on 
c          whether or not matrix values are provided.
c          Note that no error message is triggered (i.e. ierr = 0)
c          on return in these cases.  It is therefore important to
c          compare the values of job on entry and return.
c
c a	 = (int pointer) the a matrix in the a, rwptr, colind 
c                        (column) storage format
c colind = (int pointer) column number of element a(i,j) in array a.
c rwptr  = (int pointer) pointer array. rwptr(i) points to the 
c                        beginning of column i, and rwptr(nrow+1) points to
c                        one position beyond the end of arrays a, colind.
c
c rhs    = (int pointer) real array of size nrow + 1 if available (see job)
c
c nrhs   = integer containing the number of right-hand sides found
c          each right hand side may be accompanied with an initial guess
c          and also the exact solution.
c
c guesol = a 2-character string indicating whether an initial guess 
c          (1-st character) and / or the exact solution (2-nd
c          character) is provided with the right hand side.
c	   If the first character of guesol is 'G' it means that an
c          an initial guess is provided for each right-hand side.
c          These are appended to the right hand-sides in the array rhs.
c	   If the second character of guesol is 'X' it means that an
c          exact solution is provided for each right-hand side.
c          These are  appended to the right hand-sides 
c          and the initial guesses (if any) in the array rhs.
c
c nrow   = number of rows in matrix
c ncol	 = number of columns in matrix 
c nnz	 = number of nonzero elements in A. This info is returned
c          even if there is not enough space in a, colind, rwptr, in order
c          to determine the minimum storage needed. 
c
c title  = (character*72) title of test matrix 
c key    = (character*8)  key of matrix 
c type   = (charatcer*3)  type of matrix.
c          For meaning of title, key and type refer to documentation 
c          Harwell/Boeing matrices.
c
c ierr   = integer used for error messages 
c         ierr  =  0 --> the matrix has been read normally. 
c         ierr  =  1 --> the array matrix could not be read 
c                        because inadequate storage provided.
c         ierr  =  2 --> the rhs/guess/sol could not be read.
c         ierr  =  3 --> a sparse rhs was specified.
c
c mneed  = integer; amount of extra storage needed to get all info.
c          If positive, gives amount of storage left over after reading
c          in the matrix and rhs data.  If negative, gives the additional
c          number of integer words needed to successfully read stuff in.
c 
c Notes:
c-------
c 1) The file inout must be open (and possibly rewound if necessary)
c    prior to calling readmt.
c 2) Refer to the documentation on the Harwell-Boeing formats
c    for details on the format assumed by readmt.
c    We summarize the format here for convenience.
c  
c    a) All lines in inout are assumed to be 80 character long.
c    b) The file consists of a header followed by the block of the 
c       column start pointers followed by the block of the
c       row indices, followed by the block of the real values and
c       finally the numerical values of the right-hand-side if a 
c       right hand side is supplied. 
c    c) The file starts by a header which contains four lines if no
c       right hand side is supplied and five lines otherwise.
c       * first line contains the title (72 characters long) followed by
c         the 8-character identifier (name of the matrix, called key)
c        [ A72,A8 ]
c       * second line contains the number of lines for each
c         of the following data blocks (4 of them) and the total number 
c         of lines excluding the header.
c        [5i4]
c       * the third line contains a three character string identifying
c         the type of matrices as they are referenced in the Harwell
c         Boeing documentation [e.g., rua, rsa,..] and the number of
c         rows, columns, nonzero entries.
c         [A3,11X,4I14]
c       * The fourth line contains the variable fortran format
c         for the following data blocks.
c         [2A16,2A20] 
c       * The fifth line is only present if right-hand-sides are 
c         supplied. It consists of three one character-strings containing
c         the storage format for the right-hand-sides 
c         ('F'= full,'M'=sparse=same as matrix), an initial guess 
c         indicator ('G' for yes), an exact solution indicator 
c         ('X' for yes), followed by the number of right-hand-sides
c         and then the number of row indices. 
c         [A3,11X,2I14] 
c     d) The three following blocks follow the header as described 
c        above.
c     e) In case the right hand-side are in sparse formats then 
c        the fourth block uses the same storage format as for the matrix
c        to describe the NRHS right hand sides provided, with a column
c        being replaced by a right hand side.
c-----------------------------------------------------------------------

      implicit none

c     ==================================================
c     Number of integer words per double precision word:
c     ==================================================
      include '../include/intdbl.inc'

c     Declarations of variables
c     =========================

      character title*72, key*8, type*3, ptrfmt*20, indfmt*20,
     1       valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol,
     1     nnz, neltvl, nrhs, nrwindx
      integer mneed, maxstr, job, iounit,ierr
      integer len, nvec, n, next, remain, ndiag, row, j

c     Pointers for work array
c     -----------------------
      integer rwptr, colind , a, b, sol, g

c     Work array
c     ----------
      integer work(maxstr)

c     Insert diagonal entries or not?
c     -------------------------------
      logical chkdiag

c=======================================================================
c        Start of Executable Statements
c=======================================================================

c     -------------------
c     Read in header info
c     -------------------
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, 
     1     rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, 
     2     valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)

c     -----------------------------------------------
c     Read in right hand side formats if rhs provided
c     -----------------------------------------------
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs , nrwindx
 11   format (a3,11x,i14,i14)


c     -----------------------
c     Anything else to read ?
c     -----------------------
      ierr = 0
      next = 1
      mneed = 0
      if (job .le. 0) return

c     ------------------------------------------------------
c     Check whether enough storage provided to read matrix.
c     Storage needed is
c        rwptr + align(rwptr) + colind + align(colind) + a,
c     where align() is a function that forces alignment on
c     double word boundaries..
c     More may be needed if diagonal entries are inserted
c     (see ndiag below).
c     ------------------------------------------------------
      remain = maxstr
      n = ncol
      mneed = n + 1      + mod(n+1,intdbl)
     &      + nnz        + mod(nnz,intdbl)
     &      + intdbl*nnz
      remain = maxstr - mneed

c     -----------------------------------------------------------
c     Note that at this point mneed only will contain the extra
c     storage needed for the matrix, exclusive of right hand side
c     data or exact solution data.
c     -----------------------------------------------------------
      if (remain .lt. 0) then
         ierr = 1
         mneed = remain
         return
      end if

c     -----------------------------------------
c     Set up pointers for rwptr, colind, and a.
c     -----------------------------------------
      rwptr = next
      colind = rwptr + n + 1

c     ------------------------------------------------------
c     Align colind so that it starts on double word boundary
c     ------------------------------------------------------
      colind = colind + mod(colind-1 ,intdbl)

c     -----------------------------------
c     Read row pointer and column numbers
c     -----------------------------------
      call rdline(iounit,ptrfmt,'int',n+1,work(rwptr),work(rwptr))
      call rdline(iounit,indfmt,'int',nnz,work(colind),work(colind))

c     ------------------------------------------------------
c     Check to see if any diagonal entries of A are missing,
c     if it is required to force diagonal entries to be
c     present.
c     ------------------------------------------------------
      ndiag = 0
      if (chkdiag) then
         do row = 1, n
            do j = work(rwptr+row-1), work(rwptr+row)-1
               if (work(colind+j-1) .eq. row) go to 20
            end do 
            ndiag = ndiag + 1
 20      continue
         end do
         nnz = ndiag + nnz
      end if 

c     ---------------------------------
c     Read values of matrix if required
c     ---------------------------------
      if (job .le. 1)  return

      if (valcrd .le. 0) then
	 job = 1
	 return
      endif

c     -------------------------
c     Now read in matrix values
c     -------------------------
      a = colind + nnz
      a = a + mod(a-1,intdbl)
      next = a

      call rdline(iounit,valfmt,'dbl',nnz-ndiag,work(a),work(a))
      next = a + intdbl*nnz

c     ------------------------
c     Read rhs if required ... 
c     ------------------------
      if (job .le. 2)  return
c     --------------------
c     ... and if available 
c     --------------------
      if ( rhscrd .le. 0) then
	 job = 2
	 return
      endif

c     ----------------------------------------------------
c     Read right-hand-side vector b, a guess if available,
c     and a true solution vector if available.
c     ----------------------------------------------------
      if (rhstyp(1:1) .eq. 'M') then 
         ierr = 3
         return
      endif
      guesol = rhstyp(2:3) 
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g')
     &              nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x')
     &              nvec=nvec+1

c     ---------------------------------------------------
c     Check number of remaining storage words, and update
c     amount used so far:
c     ---------------------------------------------------
      remain = remain - intdbl*nvec*nrow*nrhs
      mneed  = mneed + intdbl*nvec*nrow*nrhs

      if (remain .lt. 0 ) then
         ierr = 2
         mneed = remain
         return
      endif

c     ---------------------
c     Read right-hand-sides
c     ---------------------
      len = nrhs*nrow
      b = next
      call rdline(iounit,rhsfmt,'dbl',len,work(b),work(b))
      next = b + intdbl*nrhs*nrow

c     ---------------------------------
c     Read initial guesses if available
c     ---------------------------------
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
         g = next
         call rdline(iounit,rhsfmt,'dbl',len,work(g),work(g))
         next = g  + intdbl*nrhs*nrow
      endif

c     ---------------------------------
c     Read exact solutions if available
c     ---------------------------------
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
         sol = next
         call rdline(iounit,rhsfmt,'dbl',len,work(sol),work(sol))
         next = sol  + intdbl*nrhs*nrow
      endif
c     
      return

c--------- End of readmt3 ----------------------------------------------

      end

c
c========================================================================        
c
      subroutine rdline(iounit,fmt,type,n,iarray,rarray)
      implicit none
c----------------------------------------------------------------------
c
c  Read in a line of input, specified by format fmt from IO unit
c  number iounit.  The number of words to read in is give by
c  n, and the kind (integer or double precision) by type.  One
c  of iarray or rarray will be used, depending on the type.  The
c  other can be a dummy array.
c
c----------------------------------------------------------------------
      integer iounit, n, iarray(n)
      integer i
      real*8 rarray(n)
      character*3 type
      character*20 fmt
      if (type .eq. 'int') then
         read(iounit,fmt) (iarray(i), i = 1, n)
      end if
      if (type .eq. 'dbl') then
         read(iounit,fmt) (rarray(i), i = 1, n)
      end if
      return
      end
c
c========================================================================        
c
      subroutine bmux (m, n, x, y, a, ja, ia) 
c-----------------------------------------------------------------------
c          y = A*x
c----------------------------------------------------------------------- 
c Multiplies a matrix times a vector using the dot product form
c A is stored in compressed sparse row storage.  This version is 
c instrumented, by adding its total time to the times() array on
c each invocation.  
c
c on entry:
c----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length m, containing the product y=Ax
c
c-----------------------------------------------------------------------
     
      implicit none

c     Arguments
c     =========
      integer m, n, ja(*), ia(*)
      real*8  x(n), y(m), a(*) 

c     Local variables
c     ===============
      real*8 t
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start Executable Statements
c=======================================================================

      tstart = mytime()

      do 100 i = 1, m

         y(i) = 0.0d0
c        ------------------------------------------------
c        Compute the inner product of row i with vector x
c        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue
c
      times(3) = times(3) + mytime() - tstart
      return

c=============== End of bmux ==========================================

      end
c
c=======================================================================
      subroutine btmux (m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     ----------------------
c     Zero out output vector
c     ----------------------
      do i = 1,n
         y(i) = 0.0
      end do

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmux ======================================

      end

c----------------------------------------------------------------------
      subroutine setconst(type,n,iarray,rarray,iconst,rconst)
c----------------------------------------------------------------------
c
c  Initialize the array of specified type to be the given constant.
c
c----------------------------------------------------------------------
      implicit none
      integer n, iarray(n), i, iconst
      real*8 rarray(n), rconst
      character*3 type
      if (type .eq. 'int') then
         do i = 1, n
            iarray(i) = iconst
         end do
      end if
      if (type .eq. 'dbl') then
         do i = 1, n
            rarray(i) = rconst
         end do
      end if
 1000 format(8(i6,3x))
 1010 format(4(d12.6,3x))
      return
c----------------------------------------------------------------------
       end
c----------------------------------------------------------------------
c----------------------------------------------------------------------
      subroutine setb(n,b)
c----------------------------------------------------------------------
c
c  Initialize the right hand side vector b
c
c----------------------------------------------------------------------
      implicit none
      integer n, i
      real*8 b(n)
      do i = 1, n
         b(i) = 1.0d0 - 2.0d0/dfloat(i)
c         b(i) = 1.0d0 
      end do 
      return
c----------------------------------------------------------------------
       end
c
c
c-------------------------------------------------------------------------
      subroutine getdia (nrow,ncol,job,a,ja,ia,len,diag,idiag,ioff)
      implicit real*8 (a-h,o-z)
      real*8 diag(*),a(*)
      integer nrow, ncol, job, len, ia(*), ja(*), idiag(*)
c-----------------------------------------------------------------------
c this subroutine extracts a given diagonal from a matrix stored in csr 
c format. the output matrix may be transformed with the diagonal removed
c from it if desired (as indicated by job.) 
c----------------------------------------------------------------------- 
c our definition of a diagonal of matrix is a vector of length nrow
c (always) which contains the elements in rows 1 to nrow of
c the matrix that are contained in the diagonal offset by ioff
c with respect to the main diagonal. if the diagonal element
c falls outside the matrix then it is defined as a zero entry.
c thus the proper definition of diag(*) with offset ioff is 
c
c     diag(k) = a(k,ioff+k) k=1,2,...,nrow
c     with elements falling outside the matrix being defined as zero.
c 
c----------------------------------------------------------------------- 
c 
c on entry:
c---------- 
c
c nrow	= integer. the row dimension of the matrix a.
c ncol	= integer. the column dimension of the matrix a.
c job   = integer. job indicator.  if job = 0 then
c         the matrix a, ja, ia, is not altered on return.
c         if job.ne.1  then getdia will remove the entries
c         collected in diag from the original matrix.
c         this is done in place.
c
c a,ja,
c    ia = matrix stored in compressed sparse row a,ja,ia,format
c ioff  = integer,containing the offset of the wanted diagonal
c	  the diagonal extracted is the one corresponding to the
c	  entries a(i,j) with j-i = ioff.
c	  thus ioff = 0 means the main diagonal
c
c on return:
c----------- 
c len   = number of nonzero elements found in diag.
c         (len .le. min(nrow,ncol-ioff)-max(1,1-ioff) + 1 )
c
c diag  = real*8 array of length nrow containing the wanted diagonal.
c	  diag contains the diagonal (a(i,j),j-i = ioff ) as defined 
c         above. 
c
c idiag = integer array of  length len, containing the poisitions 
c         in the original arrays a and ja of the diagonal elements
c         collected in diag. a zero entry in idiag(i) means that 
c         there was no entry found in row i belonging to the diagonal.
c         
c a, ja,
c    ia = if job .ne. 0 the matrix is unchanged. otherwise the nonzero
c         diagonal entries collected in diag are removed from the 
c         matrix. the structure is modified since the diagonal elements
c	  are removed from a,ja,ia. thus, the  returned matrix will 
c         have len fewer elements if the diagonal is full.
c----------------------------------------------------------------------c
c           y. saad, sep. 21 1989 - modified and tested may 9, 1990.   c
c----------------------------------------------------------------------c
c local variables
      integer istart, max, iend, i, kold, k, kdiag, ko
c
      istart = max(0,-ioff)
      iend = min0(nrow,ncol-ioff)
      len = 0
      do 1 i=1,nrow
         idiag(i) = 0
	 diag(i) = 0.0d0 
 1    continue
c     
c     extract  diagonal elements
c
      do 6 i=istart+1, iend
         do 51 k= ia(i),ia(i+1) -1
            if (ja(k)-i .eq. ioff) then
               diag(i)= a(k)
               idiag(i) = k
               len = len+1
               goto 6
            endif
 51      continue
 6    continue
      if (job .eq. 0 .or. len .eq.0) return
c------------- rewind the structure ------------------------------------ 
      ko = 0
      do  7 i=istart+1,iend
         kold = ko
         kdiag = idiag(i) 
         if (kdiag .eq. 0) goto 7
         do 71 k= ia(i), ia(i+1)-1 
            if (ja(k) .eq. kdiag) goto 71
            ko = ko+1
            a(ko) = a(k)
            ja(ko) = ja(k)
 71      continue
         ia(i) = kold+1
 7    continue
c
c redefine ia(nrow+1)
c
      ia(nrow+1) = ko+1
      return
c------------end-of-getdia----------------------------------------------
c-----------------------------------------------------------------------
      end


       subroutine unpack(reltv,  rpcprm    ,pcmeth  ,
     1                   ipcprm   ,solmeth ,krylov  ,
     2                   tol     ,maxits  , redo,
     3                   iunit   ,
     4                   instlvl , scale,
     5                   title, key, solver, precond)
c========================================================================        
c
c   Read in parameters that define stopping tolerances, methods to be
c   used.  Unpack information into pcmeth, ipcprm, etc.  Stop if
c   solmeth = 0.
c
c     Randall Bramley
c     Department of Computer Science
c     Indiana University
c     email: bramley@cs.indiana.edu
c     Wed Jan 26 13:23:46 EST 1994
c
c========================================================================        

      implicit none

      integer pcmeth, ipcprm, solmeth
      integer krylov, maxits
      integer iunit(20)
      real*8 rpcprm, tol
      logical reltv, results, redo
      integer instlvl, scale
      character title*72, key*8
      character solver*5, precond*4

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      results = (iunit(1) .ne. 0)
      read(iunit(2),*) solmeth

c     -----------------------------------------------------------
c     Sentinel value of solmeth = 0 is used for stopping the run.
c     -----------------------------------------------------------
      if (solmeth .le. 0) stop

      read(iunit(2),*) krylov
      read(iunit(2),*) pcmeth
      read(iunit(2),*) ipcprm
      read(iunit(2),*) rpcprm

      if (pcmeth .eq. 7) then
         if (rpcprm .gt. 1.0d0) ipcprm = int(rpcprm)
         if (rpcprm .lt. 1.0d0 .and. rpcprm .gt. 0.0d0) ipcprm = -2
         if (rpcprm .eq. 0.0d0) ipcprm = 0
         if (rpcprm .eq. 1.0d0) ipcprm = -1
      end if

      read(iunit(2),*) tol, reltv
      read(iunit(2),*) maxits
      read(iunit(2),*) instlvl
      read(iunit(2),*) scale
      read(iunit(2),*) redo

c     ------------------------------------
c     Check input parameters for validity:
c     ------------------------------------
      if (tol .lt. 0.0) then
         if (results) write(iunit(1),4040) tol
         tol = 1.0d-6
      end if
      if (rpcprm .lt. 0.0) then
         if (results) write(iunit(1),4045) rpcprm
         stop
      end if
      if (solmeth.lt. 0 .or. solmeth .gt. 14) then
         if (results) write(iunit(1),4010) solmeth
         stop
      end if
      if (pcmeth .lt. 0 .or. pcmeth .gt. 7) then
         if (results) write(iunit(1),4020) pcmeth
         stop
      end if
      if (scale .lt. 0 .or. scale .gt. 4) then
         if (results) write(iunit(1),4050) scale
         stop
      end if
c
c  Set up character strings identifying the solver/precond pair:
c
      call strings(solmeth, solver, pcmeth, precond)

c       =================
c       Format Statements
c       =================
 4010  format(' *** Invalid solution method parameter: solmeth = ',i5)
 4020  format(' *** Invalid pc method parameter: pcmeth = ',i5)
 4050  format(' *** Invalid scaling method parameter: scale = ',i2)
 4040  format(' *** Invalid tolerance: tol = ',1pe12.5,/,
     &        '     tol being reset to 1.0e-6')
 4045  format(' *** Invalid parameter: s = ',1pe12.5)

	return

c======================= End of unpack =================================

	end

c======================================================================
      subroutine prtval(iunit,type,n,iarray,rarray)
c----------------------------------------------------------------------
c
c  Print out n values of the array iarray or rarray, as
c  specified by the character array type.
c
c----------------------------------------------------------------------
      implicit none
      integer n, iarray(n), i, iunit
      real*8 rarray(n)
      character*3 type
      if (type .eq. 'int') then
         write(iunit,1000) (iarray(i), i = 1, n)
      end if
      if (type .eq. 'dbl') then
         write(iunit,1010) (rarray(i), i = 1, n)
      end if
c 1000 format(8(i7,3x))
c 1010 format(4(d18.12,3x))
 1000 format(1(i7,3x))
 1010 format(1(d24.16,3x))
      return
c----------------------------------------------------------------------
       end
c----------------------------------------------------------------------

       subroutine setlog( n, lvec, lval)
       integer n, k
       logical lvec(n), lval
       do k = 1, n
          lvec(k) = lval
       end do
       return
       end

c
c=======================================================================
c
      DOUBLE PRECISION FUNCTION DLAMCH( CMACH )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      CHARACTER          CMACH
*     ..
*
*  Purpose
*  =======
*
*  DLAMCH determines double precision machine parameters.
*
*  Arguments
*  =========
*
*  CMACH   (input) CHARACTER*1
*          Specifies the value to be returned by DLAMCH:
*          = 'E' or 'e',   DLAMCH := eps
*          = 'S' or 's ,   DLAMCH := sfmin
*          = 'B' or 'b',   DLAMCH := base
*          = 'P' or 'p',   DLAMCH := eps*base
*          = 'N' or 'n',   DLAMCH := t
*          = 'R' or 'r',   DLAMCH := rnd
*          = 'M' or 'm',   DLAMCH := emin
*          = 'U' or 'u',   DLAMCH := rmin
*          = 'L' or 'l',   DLAMCH := emax
*          = 'O' or 'o',   DLAMCH := rmax
*
*          where
*
*          eps   = relative machine precision
*          sfmin = safe minimum, such that 1/sfmin does not overflow
*          base  = base of the machine
*          prec  = eps*base
*          t     = number of (base) digits in the mantissa
*          rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
*          emin  = minimum exponent before (gradual) underflow
*          rmin  = underflow threshold - base**(emin-1)
*          emax  = largest exponent before overflow
*          rmax  = overflow threshold  - (base**emax)*(1-eps)
*
* =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE, ZERO
      PARAMETER          ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            FIRST, LRND
      INTEGER            BETA, IMAX, IMIN, IT
      DOUBLE PRECISION   BASE, EMAX, EMIN, EPS, PREC, RMACH, RMAX, RMIN,
     $                   RND, SFMIN, SMALL, T
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     ..
*     .. External Subroutines ..
      EXTERNAL           DLAMC2
*     ..
*     .. Save statement ..
      SAVE               FIRST, EPS, SFMIN, BASE, T, RND, EMIN, RMIN,
     $                   EMAX, RMAX, PREC
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         FIRST = .FALSE.
         CALL DLAMC2( BETA, IT, LRND, EPS, IMIN, RMIN, IMAX, RMAX )
         BASE = BETA
         T = IT
         IF( LRND ) THEN
            RND = ONE
            EPS = ( BASE**( 1-IT ) ) / 2
         ELSE
            RND = ZERO
            EPS = BASE**( 1-IT )
         END IF
         PREC = EPS*BASE
         EMIN = IMIN
         EMAX = IMAX
         SFMIN = RMIN
         SMALL = ONE / RMAX
         IF( SMALL.GE.SFMIN ) THEN
*
*           Use SMALL plus a bit, to avoid the possibility of rounding
*           causing overflow when computing  1/sfmin.
*
            SFMIN = SMALL*( ONE+EPS )
         END IF
      END IF
*
      IF( LSAME( CMACH, 'E' ) ) THEN
         RMACH = EPS
      ELSE IF( LSAME( CMACH, 'S' ) ) THEN
         RMACH = SFMIN
      ELSE IF( LSAME( CMACH, 'B' ) ) THEN
         RMACH = BASE
      ELSE IF( LSAME( CMACH, 'P' ) ) THEN
         RMACH = PREC
      ELSE IF( LSAME( CMACH, 'N' ) ) THEN
         RMACH = T
      ELSE IF( LSAME( CMACH, 'R' ) ) THEN
         RMACH = RND
      ELSE IF( LSAME( CMACH, 'M' ) ) THEN
         RMACH = EMIN
      ELSE IF( LSAME( CMACH, 'U' ) ) THEN
         RMACH = RMIN
      ELSE IF( LSAME( CMACH, 'L' ) ) THEN
         RMACH = EMAX
      ELSE IF( LSAME( CMACH, 'O' ) ) THEN
         RMACH = RMAX
      END IF
*
      DLAMCH = RMACH
      RETURN
*
*     End of DLAMCH
*
      END
*
************************************************************************
*
      SUBROUTINE DLAMC1( BETA, T, RND, IEEE1 )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            IEEE1, RND
      INTEGER            BETA, T
*     ..
*
*  Purpose
*  =======
*
*  DLAMC1 determines the machine parameters given by BETA, T, RND, and
*  IEEE1.
*
*  Arguments
*  =========
*
*  BETA    (output) INTEGER
*          The base of the machine.
*
*  T       (output) INTEGER
*          The number of ( BETA ) digits in the mantissa.
*
*  RND     (output) LOGICAL
*          Specifies whether proper rounding  ( RND = .TRUE. )  or
*          chopping  ( RND = .FALSE. )  occurs in addition. This may not
*          be a reliable guide to the way in which the machine performs
*          its arithmetic.
*
*  IEEE1   (output) LOGICAL
*          Specifies whether rounding appears to be done in the IEEE
*          'round to nearest' style.
*
*  Further Details
*  ===============
*
*  The routine is based on the routine  ENVRON  by Malcolm and
*  incorporates suggestions by Gentleman and Marovich. See
*
*     Malcolm M. A. (1972) Algorithms to reveal properties of
*        floating-point arithmetic. Comms. of the ACM, 15, 949-951.
*
*     Gentleman W. M. and Marovich S. B. (1974) More on algorithms
*        that reveal properties of floating point arithmetic units.
*        Comms. of the ACM, 17, 276-277.
*
* =====================================================================
*
*     .. Local Scalars ..
      LOGICAL            FIRST, LIEEE1, LRND
      INTEGER            LBETA, LT
      DOUBLE PRECISION   A, B, C, F, ONE, QTR, SAVEC, T1, T2
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Save statement ..
      SAVE               FIRST, LIEEE1, LBETA, LRND, LT
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         FIRST = .FALSE.
         ONE = 1
*
*        LBETA,  LIEEE1,  LT and  LRND  are the  local values  of  BETA,
*        IEEE1, T and RND.
*
*        Throughout this routine  we use the function  DLAMC3  to ensure
*        that relevant values are  stored and not held in registers,  or
*        are not affected by optimizers.
*
*        Compute  a = 2.0**m  with the  smallest positive integer m such
*        that
*
*           fl( a + 1.0 ) = a.
*
         A = 1
         C = 1
*
*+       WHILE( C.EQ.ONE )LOOP
   10    CONTINUE
         IF( C.EQ.ONE ) THEN
            A = 2*A
            C = DLAMC3( A, ONE )
            C = DLAMC3( C, -A )
            GO TO 10
         END IF
*+       END WHILE
*
*        Now compute  b = 2.0**m  with the smallest positive integer m
*        such that
*
*           fl( a + b ) .gt. a.
*
         B = 1
         C = DLAMC3( A, B )
*
*+       WHILE( C.EQ.A )LOOP
   20    CONTINUE
         IF( C.EQ.A ) THEN
            B = 2*B
            C = DLAMC3( A, B )
            GO TO 20
         END IF
*+       END WHILE
*
*        Now compute the base.  a and c  are neighbouring floating point
*        numbers  in the  interval  ( beta**t, beta**( t + 1 ) )  and so
*        their difference is beta. Adding 0.25 to c is to ensure that it
*        is truncated to beta and not ( beta - 1 ).
*
         QTR = ONE / 4
         SAVEC = C
         C = DLAMC3( C, -A )
         LBETA = C + QTR
*
*        Now determine whether rounding or chopping occurs,  by adding a
*        bit  less  than  beta/2  and a  bit  more  than  beta/2  to  a.
*
         B = LBETA
         F = DLAMC3( B / 2, -B / 100 )
         C = DLAMC3( F, A )
         IF( C.EQ.A ) THEN
            LRND = .TRUE.
         ELSE
            LRND = .FALSE.
         END IF
         F = DLAMC3( B / 2, B / 100 )
         C = DLAMC3( F, A )
         IF( ( LRND ) .AND. ( C.EQ.A ) )
     $      LRND = .FALSE.
*
*        Try and decide whether rounding is done in the  IEEE  'round to
*        nearest' style. B/2 is half a unit in the last place of the two
*        numbers A and SAVEC. Furthermore, A is even, i.e. has last  bit
*        zero, and SAVEC is odd. Thus adding B/2 to A should not  change
*        A, but adding B/2 to SAVEC should change SAVEC.
*
         T1 = DLAMC3( B / 2, A )
         T2 = DLAMC3( B / 2, SAVEC )
         LIEEE1 = ( T1.EQ.A ) .AND. ( T2.GT.SAVEC ) .AND. LRND
*
*        Now find  the  mantissa, t.  It should  be the  integer part of
*        log to the base beta of a,  however it is safer to determine  t
*        by powering.  So we find t as the smallest positive integer for
*        which
*
*           fl( beta**t + 1.0 ) = 1.0.
*
         LT = 0
         A = 1
         C = 1
*
*+       WHILE( C.EQ.ONE )LOOP
   30    CONTINUE
         IF( C.EQ.ONE ) THEN
            LT = LT + 1
            A = A*LBETA
            C = DLAMC3( A, ONE )
            C = DLAMC3( C, -A )
            GO TO 30
         END IF
*+       END WHILE
*
      END IF
*
      BETA = LBETA
      T = LT
      RND = LRND
      IEEE1 = LIEEE1
      RETURN
*
*     End of DLAMC1
*
      END
*
************************************************************************
*
      SUBROUTINE DLAMC2( BETA, T, RND, EPS, EMIN, RMIN, EMAX, RMAX )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            RND
      INTEGER            BETA, EMAX, EMIN, T
      DOUBLE PRECISION   EPS, RMAX, RMIN
*     ..
*
*  Purpose
*  =======
*
*  DLAMC2 determines the machine parameters specified in its argument
*  list.
*
*  Arguments
*  =========
*
*  BETA    (output) INTEGER
*          The base of the machine.
*
*  T       (output) INTEGER
*          The number of ( BETA ) digits in the mantissa.
*
*  RND     (output) LOGICAL
*          Specifies whether proper rounding  ( RND = .TRUE. )  or
*          chopping  ( RND = .FALSE. )  occurs in addition. This may not
*          be a reliable guide to the way in which the machine performs
*          its arithmetic.
*
*  EPS     (output) DOUBLE PRECISION
*          The smallest positive number such that
*
*             fl( 1.0 - EPS ) .LT. 1.0,
*
*          where fl denotes the computed value.
*
*  EMIN    (output) INTEGER
*          The minimum exponent before (gradual) underflow occurs.
*
*  RMIN    (output) DOUBLE PRECISION
*          The smallest normalized number for the machine, given by
*          BASE**( EMIN - 1 ), where  BASE  is the floating point value
*          of BETA.
*
*  EMAX    (output) INTEGER
*          The maximum exponent before overflow occurs.
*
*  RMAX    (output) DOUBLE PRECISION
*          The largest positive number for the machine, given by
*          BASE**EMAX * ( 1 - EPS ), where  BASE  is the floating point
*          value of BETA.
*
*  Further Details
*  ===============
*
*  The computation of  EPS  is based on a routine PARANOIA by
*  W. Kahan of the University of California at Berkeley.
*
* =====================================================================
*
*     .. Local Scalars ..
      LOGICAL            FIRST, IEEE, IWARN, LIEEE1, LRND
      INTEGER            GNMIN, GPMIN, I, LBETA, LEMAX, LEMIN, LT,
     $                   NGNMIN, NGPMIN
      DOUBLE PRECISION   A, B, C, HALF, LEPS, LRMAX, LRMIN, ONE, RBASE,
     $                   SIXTH, SMALL, THIRD, TWO, ZERO
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. External Subroutines ..
      EXTERNAL           DLAMC1, DLAMC4, DLAMC5
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          ABS, MAX, MIN
*     ..
*     .. Save statement ..
      SAVE               FIRST, IWARN, LBETA, LEMAX, LEMIN, LEPS, LRMAX,
     $                   LRMIN, LT
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. / , IWARN / .FALSE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         FIRST = .FALSE.
         ZERO = 0
         ONE = 1
         TWO = 2
*
*        LBETA, LT, LRND, LEPS, LEMIN and LRMIN  are the local values of
*        BETA, T, RND, EPS, EMIN and RMIN.
*
*        Throughout this routine  we use the function  DLAMC3  to ensure
*        that relevant values are stored  and not held in registers,  or
*        are not affected by optimizers.
*
*        DLAMC1 returns the parameters  LBETA, LT, LRND and LIEEE1.
*
         CALL DLAMC1( LBETA, LT, LRND, LIEEE1 )
*
*        Start to find EPS.
*
         B = LBETA
         A = B**( -LT )
         LEPS = A
*
*        Try some tricks to see whether or not this is the correct  EPS.
*
         B = TWO / 3
         HALF = ONE / 2
         SIXTH = DLAMC3( B, -HALF )
         THIRD = DLAMC3( SIXTH, SIXTH )
         B = DLAMC3( THIRD, -HALF )
         B = DLAMC3( B, SIXTH )
         B = ABS( B )
         IF( B.LT.LEPS )
     $      B = LEPS
*
         LEPS = 1
*
*+       WHILE( ( LEPS.GT.B ).AND.( B.GT.ZERO ) )LOOP
   10    CONTINUE
         IF( ( LEPS.GT.B ) .AND. ( B.GT.ZERO ) ) THEN
            LEPS = B
            C = DLAMC3( HALF*LEPS, ( TWO**5 )*( LEPS**2 ) )
            C = DLAMC3( HALF, -C )
            B = DLAMC3( HALF, C )
            C = DLAMC3( HALF, -B )
            B = DLAMC3( HALF, C )
            GO TO 10
         END IF
*+       END WHILE
*
         IF( A.LT.LEPS )
     $      LEPS = A
*
*        Computation of EPS complete.
*
*        Now find  EMIN.  Let A = + or - 1, and + or - (1 + BASE**(-3)).
*        Keep dividing  A by BETA until (gradual) underflow occurs. This
*        is detected when we cannot recover the previous A.
*
         RBASE = ONE / LBETA
         SMALL = ONE
         DO 20 I = 1, 3
            SMALL = DLAMC3( SMALL*RBASE, ZERO )
   20    CONTINUE
         A = DLAMC3( ONE, SMALL )
         CALL DLAMC4( NGPMIN, ONE, LBETA )
         CALL DLAMC4( NGNMIN, -ONE, LBETA )
         CALL DLAMC4( GPMIN, A, LBETA )
         CALL DLAMC4( GNMIN, -A, LBETA )
         IEEE = .FALSE.
*
         IF( ( NGPMIN.EQ.NGNMIN ) .AND. ( GPMIN.EQ.GNMIN ) ) THEN
            IF( NGPMIN.EQ.GPMIN ) THEN
               LEMIN = NGPMIN
*            ( Non twos-complement machines, no gradual underflow;
*              e.g.,  VAX )
            ELSE IF( ( GPMIN-NGPMIN ).EQ.3 ) THEN
               LEMIN = NGPMIN - 1 + LT
               IEEE = .TRUE.
*            ( Non twos-complement machines, with gradual underflow;
*              e.g., IEEE standard followers )
            ELSE
               LEMIN = MIN( NGPMIN, GPMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE IF( ( NGPMIN.EQ.GPMIN ) .AND. ( NGNMIN.EQ.GNMIN ) ) THEN
            IF( ABS( NGPMIN-NGNMIN ).EQ.1 ) THEN
               LEMIN = MAX( NGPMIN, NGNMIN )
*            ( Twos-complement machines, no gradual underflow;
*              e.g., CYBER 205 )
            ELSE
               LEMIN = MIN( NGPMIN, NGNMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE IF( ( ABS( NGPMIN-NGNMIN ).EQ.1 ) .AND.
     $            ( GPMIN.EQ.GNMIN ) ) THEN
            IF( ( GPMIN-MIN( NGPMIN, NGNMIN ) ).EQ.3 ) THEN
               LEMIN = MAX( NGPMIN, NGNMIN ) - 1 + LT
*            ( Twos-complement machines with gradual underflow;
*              no known machine )
            ELSE
               LEMIN = MIN( NGPMIN, NGNMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE
            LEMIN = MIN( NGPMIN, NGNMIN, GPMIN, GNMIN )
*         ( A guess; no known machine )
            IWARN = .TRUE.
         END IF
***
* Comment out this if block if EMIN is ok
         IF( IWARN ) THEN
            FIRST = .TRUE.
            WRITE( 6, FMT = 9999 )LEMIN
         END IF
***
*
*        Assume IEEE arithmetic if we found denormalised  numbers above,
*        or if arithmetic seems to round in the  IEEE style,  determined
*        in routine DLAMC1. A true IEEE machine should have both  things
*        true; however, faulty machines may have one or the other.
*
         IEEE = IEEE .OR. LIEEE1
*
*        Compute  RMIN by successive division by  BETA. We could compute
*        RMIN as BASE**( EMIN - 1 ),  but some machines underflow during
*        this computation.
*
         LRMIN = 1
         DO 30 I = 1, 1 - LEMIN
            LRMIN = DLAMC3( LRMIN*RBASE, ZERO )
   30    CONTINUE
*
*        Finally, call DLAMC5 to compute EMAX and RMAX.
*
         CALL DLAMC5( LBETA, LT, LEMIN, IEEE, LEMAX, LRMAX )
      END IF
*
      BETA = LBETA
      T = LT
      RND = LRND
      EPS = LEPS
      EMIN = LEMIN
      RMIN = LRMIN
      EMAX = LEMAX
      RMAX = LRMAX
*
      RETURN
*
 9999 FORMAT( / / ' WARNING. The value EMIN may be incorrect:-',
     $      '  EMIN = ', I8, /
     $      ' If, after inspection, the value EMIN looks',
     $      ' acceptable please comment out ',
     $      / ' the IF block as marked within the code of routine',
     $      ' DLAMC2,', / ' otherwise supply EMIN explicitly.', / )
*
*     End of DLAMC2
*
      END
*
************************************************************************
*
      DOUBLE PRECISION FUNCTION DLAMC3( A, B )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      DOUBLE PRECISION   A, B
*     ..
*
*  Purpose
*  =======
*
*  DLAMC3  is intended to force  A  and  B  to be stored prior to doing
*  the addition of  A  and  B ,  for use in situations where optimizers
*  might hold one of these in a register.
*
*  Arguments
*  =========
*
*  A, B    (input) DOUBLE PRECISION
*          The values A and B.
*
* =====================================================================
*
*     .. Executable Statements ..
*
      DLAMC3 = A + B
*
      RETURN
*
*     End of DLAMC3
*
      END
*
************************************************************************
*
      SUBROUTINE DLAMC4( EMIN, START, BASE )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      INTEGER            BASE, EMIN
      DOUBLE PRECISION   START
*     ..
*
*  Purpose
*  =======
*
*  DLAMC4 is a service routine for DLAMC2.
*
*  Arguments
*  =========
*
*  EMIN    (output) EMIN
*          The minimum exponent before (gradual) underflow, computed by
*          setting A = START and dividing by BASE until the previous A
*          can not be recovered.
*
*  START   (input) DOUBLE PRECISION
*          The starting point for determining EMIN.
*
*  BASE    (input) INTEGER
*          The base of the machine.
*
* =====================================================================
*
*     .. Local Scalars ..
      INTEGER            I
      DOUBLE PRECISION   A, B1, B2, C1, C2, D1, D2, ONE, RBASE, ZERO
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Executable Statements ..
*
      A = START
      ONE = 1
      RBASE = ONE / BASE
      ZERO = 0
      EMIN = 1
      B1 = DLAMC3( A*RBASE, ZERO )
      C1 = A
      C2 = A
      D1 = A
      D2 = A
*+    WHILE( ( C1.EQ.A ).AND.( C2.EQ.A ).AND.
*    $       ( D1.EQ.A ).AND.( D2.EQ.A )      )LOOP
   10 CONTINUE
      IF( ( C1.EQ.A ) .AND. ( C2.EQ.A ) .AND. ( D1.EQ.A ) .AND.
     $    ( D2.EQ.A ) ) THEN
         EMIN = EMIN - 1
         A = B1
         B1 = DLAMC3( A / BASE, ZERO )
         C1 = DLAMC3( B1*BASE, ZERO )
         D1 = ZERO
         DO 20 I = 1, BASE
            D1 = D1 + B1
   20    CONTINUE
         B2 = DLAMC3( A*RBASE, ZERO )
         C2 = DLAMC3( B2 / RBASE, ZERO )
         D2 = ZERO
         DO 30 I = 1, BASE
            D2 = D2 + B2
   30    CONTINUE
         GO TO 10
      END IF
*+    END WHILE
*
      RETURN
*
*     End of DLAMC4
*
      END
*
************************************************************************
*
      SUBROUTINE DLAMC5( BETA, P, EMIN, IEEE, EMAX, RMAX )
*
*  -- LAPACK auxiliary routine (version 1.1) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            IEEE
      INTEGER            BETA, EMAX, EMIN, P
      DOUBLE PRECISION   RMAX
*     ..
*
*  Purpose
*  =======
*
*  DLAMC5 attempts to compute RMAX, the largest machine floating-point
*  number, without overflow.  It assumes that EMAX + abs(EMIN) sum
*  approximately to a power of 2.  It will fail on machines where this
*  assumption does not hold, for example, the Cyber 205 (EMIN = -28625,
*  EMAX = 28718).  It will also fail if the value supplied for EMIN is
*  too large (i.e. too close to zero), probably with overflow.
*
*  Arguments
*  =========
*
*  BETA    (input) INTEGER
*          The base of floating-point arithmetic.
*
*  P       (input) INTEGER
*          The number of base BETA digits in the mantissa of a
*          floating-point value.
*
*  EMIN    (input) INTEGER
*          The minimum exponent before (gradual) underflow.
*
*  IEEE    (input) LOGICAL
*          A logical flag specifying whether or not the arithmetic
*          system is thought to comply with the IEEE standard.
*
*  EMAX    (output) INTEGER
*          The largest exponent before overflow
*
*  RMAX    (output) DOUBLE PRECISION
*          The largest machine floating-point number.
*
* =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ZERO, ONE
      PARAMETER          ( ZERO = 0.0D0, ONE = 1.0D0 )
*     ..
*     .. Local Scalars ..
      INTEGER            EXBITS, EXPSUM, I, LEXP, NBITS, TRY, UEXP
      DOUBLE PRECISION   OLDY, RECBAS, Y, Z
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MOD
*     ..
*     .. Executable Statements ..
*
*     First compute LEXP and UEXP, two powers of 2 that bound
*     abs(EMIN). We then assume that EMAX + abs(EMIN) will sum
*     approximately to the bound that is closest to abs(EMIN).
*     (EMAX is the exponent of the required number RMAX).
*
      LEXP = 1
      EXBITS = 1
   10 CONTINUE
      TRY = LEXP*2
      IF( TRY.LE.( -EMIN ) ) THEN
         LEXP = TRY
         EXBITS = EXBITS + 1
         GO TO 10
      END IF
      IF( LEXP.EQ.-EMIN ) THEN
         UEXP = LEXP
      ELSE
         UEXP = TRY
         EXBITS = EXBITS + 1
      END IF
*
*     Now -LEXP is less than or equal to EMIN, and -UEXP is greater
*     than or equal to EMIN. EXBITS is the number of bits needed to
*     store the exponent.
*
      IF( ( UEXP+EMIN ).GT.( -LEXP-EMIN ) ) THEN
         EXPSUM = 2*LEXP
      ELSE
         EXPSUM = 2*UEXP
      END IF
*
*     EXPSUM is the exponent range, approximately equal to
*     EMAX - EMIN + 1 .
*
      EMAX = EXPSUM + EMIN - 1
      NBITS = 1 + EXBITS + P
*
*     NBITS is the total number of bits needed to store a
*     floating-point number.
*
      IF( ( MOD( NBITS, 2 ).EQ.1 ) .AND. ( BETA.EQ.2 ) ) THEN
*
*        Either there are an odd number of bits used to store a
*        floating-point number, which is unlikely, or some bits are
*        not used in the representation of numbers, which is possible,
*        (e.g. Cray machines) or the mantissa has an implicit bit,
*        (e.g. IEEE machines, Dec Vax machines), which is perhaps the
*        most likely. We have to assume the last alternative.
*        If this is true, then we need to reduce EMAX by one because
*        there must be some way of representing zero in an implicit-bit
*        system. On machines like Cray, we are reducing EMAX by one
*        unnecessarily.
*
         EMAX = EMAX - 1
      END IF
*
      IF( IEEE ) THEN
*
*        Assume we are on an IEEE machine which reserves one exponent
*        for infinity and NaN.
*
         EMAX = EMAX - 1
      END IF
*
*     Now create RMAX, the largest machine number, which should
*     be equal to (1.0 - BETA**(-P)) * BETA**EMAX .
*
*     First compute 1.0 - BETA**(-P), being careful that the
*     result is less than 1.0 .
*
      RECBAS = ONE / BETA
      Z = BETA - ONE
      Y = ZERO
      DO 20 I = 1, P
         Z = Z*RECBAS
         IF( Y.LT.ONE )
     $      OLDY = Y
         Y = DLAMC3( Y, Z )
   20 CONTINUE
      IF( Y.GE.ONE )
     $   Y = OLDY
*
*     Now multiply by BETA**EMAX to get RMAX.
*
      DO 30 I = 1, EMAX
         Y = DLAMC3( Y*BETA, ZERO )
   30 CONTINUE
*
      RMAX = Y
      RETURN
*
*     End of DLAMC5
*
      END
c
c----------------------------------------------------------------------
c=============================================================================

c
      subroutine bmuxp(m, n, x, y, a, ja, ia) 
c-----------------------------------------------------------------------
c          y = y + A*x
c----------------------------------------------------------------------- 
c Multiplies a matrix times a vector using the dot product form
c A is stored in compressed sparse row storage.  This version is 
c instrumented, by adding its total time to the times() array on
c each invocation.  
c
c on entry:
c----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length m, containing the product y=Ax
c
c-----------------------------------------------------------------------
     
      implicit none

c     Arguments
c     =========
      integer m, n, ja(*), ia(*)
      real*8  x(n), y(m), a(*) 

c     Local variables
c     ===============
      real*8 t
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start Executable Statements
c=======================================================================

      tstart = mytime()

      do 100 i = 1, m

c        ------------------------------------------------
c        Compute the inner product of row i with vector x
c        ------------------------------------------------
         do k = ia(i), ia(i+1)-1 
            y(i) = y(i) + a(k)*x(ja(k))
         end do

 100  continue
c
      times(3) = times(3) + mytime() - tstart
      return

c=============== End of bmuxp =========================================

      end
c
c=======================================================================
      subroutine btmuxp(m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  y + A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.  This version
c updates with an add to y.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmuxp ======================================

      end
 
c=======================================================================
      subroutine btmuxm(m, n, x, y, a, ja, ia)
c-----------------------------------------------------------------------
c         y =  y - A' * x
c
c Multiplies the transpose of a matrix times a vector when the original
c matrix is stored in compressed sparse row storage. Can also be
c used as the product of a matrix by a vector when the original
c matrix is stored in the compressed sparse column format.  This version
c updates with a subtraction from y.
c-----------------------------------------------------------------------
c
c on entry:
c ----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the row dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c -----------
c y     = real array of length equal to column dimmension of the A matrix, 
c	  containing the product y=transp(A)*x
c
c-----------------------------------------------------------------------

      implicit none

c     Arguments
c     =========
      integer m, n, ia(*), ja(*)
      real*8 x(m), y(n), a(*)

c     Local variables
c     ===============
      integer i, k

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start of Executable Statements
c=======================================================================

      tstart = mytime()

c     -------------------------------------------------
c     Loop over the rows, using saxpy form of multiply.
c     -------------------------------------------------
      do i = 1, m
         do k = ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) - x(i)*a(k)
         end do
      end do

      times(4) = times(4) + mytime() - tstart
      return

c=================== End of btmuxm ======================================

      end

c-----------------------------------------------------------------------
      subroutine readunf2(maxstr,nrow,ncol,nnz,a,ja,ia,iounit,
     1                  work,mneed,ierr,nrhs,rhs,title,key,type,digits)
c-----------------------------------------------------------------------
c This subroutine reads a matix stored in binary CSR format.  This differs
c from readunf in that it will read the matrix into a work vector using
c the memory manager package.
c-----------------------------------------------------------------------
c On entry:
c  maxstr: the maximum integer words allocated in work array
c  iounit: the I/O unit that opened for reading.
c  nrhs  : 0 if no right hand side vector is to be read, nonzero
c          otherwise.  nrhs is reset to zero if no rhs vector is
c          available to be read.  This is not an error condition, so
c          if you request a rhs to be read, be sure to check nrhs on
c          return to see if one was available.
c  digits: number of significant digits for representing floats.
c On return:
c    nrow: the actual number of rows in array.
c    ncol: the actual number of cols in array.
c     nnz: the actual number of non-zero entries.
c    nrhs: the actual number of right hand side vectors read (0 or 1)
c ia,ja,a: pointers in work array to where corr arrays go.
c    ierr: a error code, it's same as that used in reaadsk
c          0 -- OK
c          1 -- error in reading iounit
c          2 -- end-of-file reached while reading data file
c          3 -- not enough storage provided; check mneed
c On error:
c     mneed contains additional storage needed if not enough was provided.
c-----------------------------------------------------------------------
c
      integer intdbl
      parameter (intdbl = 2)
      integer nrow,ncol, iounit, nnz, k, digits
      integer work(maxstr), maxstr, remain
      integer ia, ja, a, rhs, next, mneed
      character title*72, key*8, type*3
c
      ierr = 0
      rewind iounit
c      
      read(unit=iounit, err=1000, end=1010) nrow, ncol
      read(unit=iounit, err=1000, end=1010) title, key, type,
     &                                       digits, nrhs

      mneed = 0
      next = 1
      ia = next
      if (nrow+1 .gt. maxstr) then
         ierr = 3
         mneed = maxstr - nrow-1
         return
      end if 
      read(unit=iounit, err=1000, end=1010) (work(ia+k),k=0,nrow)
c
      nnz = work(ia+nrow) - 1
      next = next + nrow + 1
      next = next + mod(next-1, intdbl)
      ja = next
      remain = maxstr - next
      if (remain .lt. (intdbl+1)*nnz) then
         ierr = 3
         mneed = remain - (intdbl+1)*nnz
         return
      end if 
      read(unit=iounit, err=1000, end=1010) (work(ja+k),k=0,nnz-1)

      next = next + nnz
      next = next + mod(next-1, intdbl)
      a = next
      remain = maxstr - next

      call rlunf(work(a), nnz, iounit, ierr)
      if (ierr .ne. 0 ) return
      if (nrhs .le. 0)  return

      next = next + intdbl*nnz
      rhs = next
      remain = maxstr - next
      if (remain .lt. intdbl*nnz) then
         ierr = 3
         mneed = remain - intdbl*nnz
         return
      end if 
      call rlunf(work(rhs), nrow, iounit, ierr)
c
      return
c
c     error handling
c
 1000 ierr = 1
      return
 1010 ierr = 2
      return
      end
c---------end of readunf2 ----------------------------------------------
      subroutine rlunf(a, m, iounit, ierr)
c
c  Read in a double precision array from iounit, in binary format.
c  ierr = 0 if all's ok
c  ierr = 1 if error found in reading
c  ierr = 2 if end reached before m values read.
c
      integer m, iounit, ierr
      real*8 a(*)
      ierr  = 0
      read(unit=iounit, err=1000, end=1010) (a(k),k=1,m)
      return
 1000 ierr = 1
      return
 1010 ierr = 2
      return
      end

c=======================================================================
      subroutine spbmux (m, n, x, y, a, ja, ia) 
c-----------------------------------------------------------------------
c          y = A*x
c----------------------------------------------------------------------- 
c Multiplies a matrix times a vector using the dot product form.
c A is stored in compressed sparse row storage.  This version is 
c instrumented, and it calls the C routine spdot.
c
c on entry:
c----------
c m     = row dimension of A
c n     = column dimension of A
c x     = real array of length equal to the column dimension of
c         the A matrix.
c a, ja,
c    ia = input matrix in compressed sparse row format.
c
c on return:
c-----------
c y     = real array of length m, containing the product y=Ax
c
c-----------------------------------------------------------------------
     
      implicit none

c     Arguments
c     =========
      integer m, n, ja(*), ia(*)
      real*8  x(n), y(m), a(*) 
      real*8  spdot
      external spdot

c     Local variables
c     ===============
      real*8 t
      integer i, k, start

c     Instrumentation variables for timing.
c     =====================================
      include '../include/heads.inc'
      real*8 mytime, tstart
      external mytime

c=======================================================================
c      Start Executable Statements
c=======================================================================

      tstart = mytime()

      do 100 i = 1, m

c        ------------------------------------------------
c        Compute the inner product of row i with vector x
c        ------------------------------------------------
         start = ia(i)
         k = ia(i+1) - start
         y(i) = spdot(k, x, a(start), ja(start))

 100  continue
c
      times(3) = times(3) + mytime() - tstart
      return

c=============== End of spbmux ==========================================

      end
c
c=======================================================================

      subroutine scala(job,A, rwptr, colind, n, b, D)
      implicit none

c----------------------------------------------------------------
c     Scale a CSR matrix to have certain properties, 
c     depending on the job parameter.
c       job = 1  ---> set max(max(A)) = 1.0
c       job = 2  ---> set max of each row to be 1.0
c       job = 3  ---> scale each column so diagonal entry is 1.0
c
c     If job > 1, a vector is returned in D that gives
c     the scaling vector used, so that
c          Anew = D*A   (job = 2)  or
c          Anew = A*D   (job = 3)
c
c     In case job < 3, the right hand side vector is also scaled.
c
c     Warning: this code assumes that A is not the zero matrix!
c       Also, if the max-norm of a row is zero when job = 2, 
c       or if the diagonal element is zero when job = 3, the
c       code sets the corresponding D element to be zero and
c       does not scale that row or column.
c----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      real*8 A(*), t
      real*8 b(n), D(n)

      integer idamax
      external idamax

c---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
         nnz = rwptr(n+1)-1
         k = idamax(nnz, A, 1)
         t = 1.0d0/A(k)
         call dscal(nnz,t,A,1)
         return
      end if 

c---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             t = abs(A(s))
             do k = s+1, rwptr(i+1)-1
               if (abs(A(k)) .gt. t) t = abs(A(k))
             end do 
             if ( t .ne. 0.0d0) then
                D(i) = 1.0d0/t
                call dscal(nnz,D(i),A(s),1)
                b(i) = D(i)*b(i)
             else
                D(i) = 0.0d0
             end if
         end do 
         return
      end if 
c---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         do i = 1, n
            D(i) = 1.0d0
            do k = rwptr(i), rwptr(i+1)-1
               if (colind(k) .eq. i .and. A(k).ne. 0.0d0) then
                  D(i) = 1.0d0/A(k)
                  go to 100
               end if
            end do
 100     continue
         end do 
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)*D(colind(i))
         end do 
         return
      end if 

      return
      end 
c
c==============================================================
c
      subroutine unscalx(n, x, Dr)
      implicit none
      integer n, k
      real*8 x(n), Dr(n)
      do k = 1, n
         x(k) = x(k)*Dr(k)
      end do
      return
      end

c-----------------------------------------------------------------------
c
      subroutine unscala(job,A, rwptr, colind, n, b, D)
      implicit none

c----------------------------------------------------------------
c     Unscale a CSR matrix to reverse effects of scala.
c----------------------------------------------------------------

      integer rwptr(*), n, colind(*), k, i, nnz
      integer s, job

      real*8 A(*), t
      real*8 b(n), D(n)

      integer idamax
      external idamax

c---------- Scaling one: max(max(A)) = 1.0 -------------
      if (job .eq. 1) then
c         nnz = rwptr(n+1)-1
c         k = idamax(nnz, A, 1)
c         t = 1.0d0/A(k)
c         call dscal(nnz,t,A,1)
c         return
      end if 

c---------- Scaling two: max(A') = 1.0 -------------
      if (job .eq. 2) then
         do i = 1, n
             s = rwptr(i)
             nnz = rwptr(i+1) - s
             t = 1.0d0/D(i)
             call dscal(nnz,t,A(s),1)
             b(i) = t*b(i)
         end do 
         return
      end if 
c---------- Scaling three: A*D has diag(A) = I --------
      if (job .eq. 3) then
         nnz = rwptr(n+1)-1
         do i = 1, nnz
            A(i) = A(i)/D(colind(i))
         end do 
         return
      end if 

      return
      end 
c
c-----------------------------------------------------------------------
